from django.core.files.images import ImageFile
from django.core.urlresolvers import reverse
from django.test import TestCase

from tint import models

import os
import re


class ViewsTestCase(TestCase):
    def setUp(self):
        image_file = open(os.path.join(os.path.dirname(__file__), 'test-image.png'), 'rb')
        self.image = models.Image.objects.create(
            image=ImageFile(image_file, name='test-image.png'),
            height=640,
            width=480,
        )

    def tearDown(self):
        models.Image.objects.all().delete()
        models.Thumbnail.objects.all().delete()

    def test_thumbnail_view(self):
        response = self.client.get(reverse('image-thumbnail', args=(1000000, 'test1')))
        self.assertEqual(response.status_code, 404)
        response = self.client.get(reverse('image-thumbnail', args=(self.image.id, 'not-valid-transformation')))
        self.assertEqual(response.status_code, 404)
        self.assertEqual(models.Thumbnail.objects.all().count(), 0)
        response = self.client.get(reverse('image-thumbnail', args=(self.image.id, 'test1')))
        self.assertEqual(response.status_code, 302)
        self.assertIsNotNone(re.search(
            'http://testserver/media/image/thumbnail/by-md5/[a-f0-9]/[a-f0-9]/[a-f0-9]{32}/test-image.png',
            response.get('location')
        ))
        self.assertEqual(models.Thumbnail.objects.all().count(), 1)



from django.http import HttpResponseNotFound
from django.shortcuts import get_object_or_404, redirect

from . import models
from .settings import TINT_TRANSFORMATIONS

def thumbnail(request, image_id, transformation):
    image = get_object_or_404(models.Image, id=image_id)
    if not transformation in TINT_TRANSFORMATIONS:
        return HttpResponseNotFound()

    return redirect(models.Thumbnail.objects.get_or_create_at_transformation(image.id, transformation))



from django import template

register = template.Library()


@register.filter
def at_transformation(image, transformation):
    return image.get_absolute_url(transformation=transformation)

try:
    from django_jinja.base import Library
    jinja_register = Library()

    @jinja_register.global_function
    def image_at_transformation(image, transformation):
        if image:
            return image.get_absolute_url(transformation=transformation)
        return None
except ImportError:
    pass



from django.conf import settings

TINT_TRANSFORMATIONS = getattr(settings, 'TINT_TRANSFORMATIONS', {})
TINT_IMAGE_PROCESSOR = getattr(settings, 'TINT_IMAGE_PROCESSOR', 'tint.imageprocs.default.DefaultImageProc')
TINT_KEEP_THUMBNAILS = getattr(settings, 'TINT_KEEP_THUMBNAILS', True)
TINT_KEEP_IMAGES = getattr(settings, 'TINT_KEEP_IMAGES', True)
TINT_IMAGE_OUTPUT_FORMAT = getattr(settings, 'TINT_IMAGE_OUTPUT_FORMAT', None)
TINT_IMAGE_OUTPUT_EXTENSION = getattr(settings, 'TINT_IMAGE_OUTPUT_EXTENSION', None)



from django.utils import unittest
from django.template import Template, Context
from django.core.files.images import ImageFile
from django.core.urlresolvers import reverse

from django_jinja.base import env

from tint import models

import os
import re


class TemplateTagsTestCase(unittest.TestCase):
    def setUp(self):
        image_file = open(os.path.join(os.path.dirname(__file__), 'test-image.png'), 'rb')
        self.image = models.Image.objects.create(
            image=ImageFile(image_file, name='test-image.png'),
            height=640,
            width=480,
        )

    def tearDown(self):
        models.Image.objects.all().delete()
        models.Thumbnail.objects.all().delete()

    def test_tint_django_templatetag(self):
        context = Context({"image": self.image})
        result = Template("{% load tint %}{{ image|at_transformation:'test1' }}").render(context)
        self.assertEqual(result, reverse('image-thumbnail', args=(self.image.id, 'test1')))
        self.image.get_absolute_url('test1', True)
        result = Template("{% load tint %}{{ image|at_transformation:'test1' }}").render(context)
        self.assertIsNotNone(re.search(
            '/media/image/thumbnail/by-md5/[a-f0-9]/[a-f0-9]/[a-f0-9]{32}/test-image.png',
            result
        ))

    def test_tint_django_jinja_global_function(self):
        context = Context({"image": self.image})
        result = env.from_string("{{ image_at_transformation(image, 'test1') }}").render(context)
        self.assertEqual(result, reverse('image-thumbnail', args=(self.image.id, 'test1')))
        self.image.get_absolute_url('test1', True)
        result = env.from_string("{{ image_at_transformation(image, 'test1') }}").render(context)
        self.assertIsNotNone(re.search(
            '/media/image/thumbnail/by-md5/[a-f0-9]/[a-f0-9]/[a-f0-9]{32}/test-image.png',
            result
        ))



import os

SECRET_KEY = "123"

DATABASES = {
    'default': {
        'ENGINE': 'django.db.backends.sqlite3', # Add 'postgresql_psycopg2', 'mysql', 'sqlite3' or 'oracle'.
        'NAME': 'test',                      # Or path to database file if using sqlite3.
        'USER': '',
        'PASSWORD': '',
        'HOST': '',                      # Empty for localhost through domain sockets or '127.0.0.1' for localhost through TCP.
        'PORT': '',                      # Set to empty string for default.
    }
}

INSTALLED_APPS = [
    'tint'
]

MEDIA_URL = "/media/"
ROOT_URLCONF = "tests.urls"
MEDIA_ROOT = os.path.join(os.path.dirname(__file__), "media")

TINT_TRANSFORMATIONS = {
  'test1': [
      {
          "action": 'fit',
          "width": 1024,
          "height": 768,
          "align": 'center',
          "valign": 'middle',
      },
  ],
}

TINT_KEEP_IMAGES = False
TINT_KEEP_THUMBNAILS = False

ALLOWED_HOSTS = ["testserver"]



from django.core.files.images import ImageFile
from django.core.management import call_command
from django.core.urlresolvers import reverse
from django.test import TestCase

from tint import models
import os
import re

call_command('syncdb', interactive=False)


class ModelsTestCase(TestCase):
    def setUp(self):
        image_file = open(os.path.join(os.path.dirname(__file__), 'test-image.png'), 'rb')
        self.image = models.Image.objects.create(
            image=ImageFile(image_file, name='test-image.png'),
            height=640,
            width=480,
        )

    def tearDown(self):
        models.Image.objects.all().delete()
        models.Thumbnail.objects.all().delete()

    def test_hashed_upload_to(self):
        self.assertIsNotNone(re.search(
            'test/[a-f0-9]/[a-f0-9]/[a-f0-9]{32}/test-image.png',
            models.hashed_upload_to("test/", self.image, 'test-image.png')
        ))

    def test_image_get_by_transformation(self):
        self.assertRaises(
            models.Thumbnail.DoesNotExist,
            lambda: self.image.get_by_transformation('test1')
        )
        self.image.get_absolute_url('test1', True),
        self.assertTrue(isinstance(self.image.get_by_transformation('test1'), models.Thumbnail))

    def test_image_get_absolute_url(self):
        self.assertIsNotNone(re.search(
            '/media/image/original/by-md5/[a-f0-9]/[a-f0-9]/[a-f0-9]{32}/test-image.png',
            self.image.get_absolute_url()
        ))
        self.assertEqual(
            self.image.get_absolute_url('test1'),
            reverse('image-thumbnail', args=(self.image.id, 'test1'))
        )
        self.assertIsNotNone(re.search(
            '/media/image/thumbnail/by-md5/[a-f0-9]/[a-f0-9]/[a-f0-9]{32}/test-image.png',
            self.image.get_absolute_url('test1', True)
        ))
        self.assertIsNotNone(re.search(
            '/media/image/thumbnail/by-md5/[a-f0-9]/[a-f0-9]/[a-f0-9]{32}/test-image.png',
            self.image.get_absolute_url('test1')
        ))
        models.Thumbnail.objects.all().delete()
        self.assertEqual(
            self.image.get_absolute_url('test1'),
            reverse('image-thumbnail', args=(self.image.id, 'test1'))
        )

    def test_image_upload_to(self):
        self.assertIsNotNone(re.search(
            'image/original/by-md5/[a-f0-9]/[a-f0-9]/[a-f0-9]{32}/test-image.png',
            models.image_upload_to(self.image, 'test-image.png')
        ))

    def test_thumbnail_upload_to(self):
        self.assertIsNotNone(re.search(
            'image/thumbnail/by-md5/[a-f0-9]/[a-f0-9]/[a-f0-9]{32}/test-image.png',
            models.thumbnail_upload_to(self.image, 'test-image.png')
        ))

    def test_thumbnail_manager_get_or_create_at_transformation(self):
        self.assertRaises(
            ValueError,
            lambda: models.Thumbnail.objects.get_or_create_at_transformation(self.image.id, "invalid-transformation")
        )

        self.assertRaises(
            models.Image.DoesNotExist,
            lambda: models.Thumbnail.objects.get_or_create_at_transformation(-1, "test1")
        )

        self.assertEqual(models.Thumbnail.objects.all().count(), 0)
        self.assertTrue(isinstance(
            models.Thumbnail.objects.get_or_create_at_transformation(self.image.id, "test1"),
            models.Thumbnail
        ))
        self.assertEqual(models.Thumbnail.objects.all().count(), 1)
        self.assertTrue(isinstance(
            models.Thumbnail.objects.get_or_create_at_transformation(self.image.id, "test1"),
            models.Thumbnail
        ))
        self.assertEqual(models.Thumbnail.objects.all().count(), 1)

    def test_thumbnail_get_absolute_url(self):
        self.image.get_absolute_url('test1', True),
        self.assertIsNotNone(re.search(
            '/media/image/thumbnail/by-md5/[a-f0-9]/[a-f0-9]/[a-f0-9]{32}/test-image.png',
            models.Thumbnail.objects.all()[0].get_absolute_url()
        ))

    def test_original_changed(self):
        self.assertEqual(models.Thumbnail.objects.all().count(), 0)
        models.Thumbnail.objects.get_or_create_at_transformation(self.image.id, "test1")
        self.assertEqual(models.Thumbnail.objects.all().count(), 1)
        self.image.save()
        self.assertEqual(models.Thumbnail.objects.all().count(), 0)



import hashlib
import os.path

from django.db import models
from django.core.files.uploadedfile import InMemoryUploadedFile
from django.core.urlresolvers import reverse
from django.dispatch import receiver
from django.utils.translation import ugettext_lazy as _

from .settings import TINT_TRANSFORMATIONS, TINT_KEEP_IMAGES, TINT_KEEP_THUMBNAILS, TINT_IMAGE_OUTPUT_EXTENSION

from .imageprocs import image_proc


def hashed_upload_to(prefix, instance, filename):
    hasher = hashlib.md5()
    instance.image.open('rb')
    for chunk in instance.image.chunks():
        hasher.update(chunk)
    hash = hasher.hexdigest()
    base, ext = os.path.splitext(filename)
    if TINT_IMAGE_OUTPUT_EXTENSION:
        ext = TINT_IMAGE_OUTPUT_EXTENSION
    return '%(prefix)s%(first)s/%(second)s/%(hash)s/%(base)s%(ext)s' % {
        'prefix': prefix,
        'first': hash[0],
        'second': hash[1],
        'hash': hash,
        'base': base,
        'ext': ext,
    }


def image_upload_to(instance, filename, **kwargs):
    return hashed_upload_to('image/original/by-md5/', instance, filename)


class Image(models.Model):
    image = models.ImageField(upload_to=image_upload_to,
            height_field='height', width_field='width',
            max_length=255)
    height = models.PositiveIntegerField(default=0, editable=False)
    width = models.PositiveIntegerField(default=0, editable=False)

    class Meta:
        ordering = ["id"]
        verbose_name = _(u"image")
        verbose_name_plural = _(u"images")

    def __unicode__(self):
        if self.image:
            return self.image.name
        return _(u"Image #{0}").format(self.id)

    def get_by_transformation(self, transformation):
        return self.thumbnail_set.get(transformation=transformation)

    def get_absolute_url(self, transformation=None, wait_creation=False):
        if not transformation:
            return self.image.url
        try:
            return self.get_by_transformation(transformation).image.url
        except Thumbnail.DoesNotExist:
            if wait_creation:
                thumbnail = Thumbnail.objects.get_or_create_at_transformation(self.id, transformation)
                return thumbnail.image.url
            return reverse('image-thumbnail', args=(self.id, transformation))


def thumbnail_upload_to(instance, filename, **kwargs):
    return hashed_upload_to('image/thumbnail/by-md5/', instance, filename)


class ThumbnailManager(models.Manager):
    def get_or_create_at_transformation(self, image_id, transformation):
        image = Image.objects.get(id=image_id)
        if not transformation in TINT_TRANSFORMATIONS:
            raise ValueError("Received unknown transformation: %s" % transformation)
        try:
            thumbnail = image.get_by_transformation(transformation)
        except Thumbnail.DoesNotExist:
            buf = image_proc.transform(image.image, *TINT_TRANSFORMATIONS[transformation])
            # and save to storage
            original_dir, original_file = os.path.split(image.image.name)
            thumb_file = InMemoryUploadedFile(buf, "image", original_file, None, buf.tell(), None)
            thumbnail, created = image.thumbnail_set.get_or_create(
                transformation=transformation,
                defaults={'image': thumb_file})
        return thumbnail


class Thumbnail(models.Model):
    original = models.ForeignKey(Image)
    image = models.ImageField(upload_to=thumbnail_upload_to,
            height_field='height', width_field='width',
            max_length=255)
    transformation = models.CharField(max_length=100)
    height = models.PositiveIntegerField(default=0, editable=False)
    width = models.PositiveIntegerField(default=0, editable=False)

    objects = ThumbnailManager()

    class Meta:
        ordering = ["id"]
        unique_together = ('original', 'transformation')
        verbose_name = _(u"thumbnail")
        verbose_name_plural = _(u"thumbnails")

    def __unicode__(self):
        if self.image:
            return self.image.name
        return _(u"Thumbnail Image #{0}").format(self.id)

    def get_absolute_url(self):
        return self.image.url


@receiver(models.signals.post_save)
def original_changed(sender, instance, created, **kwargs):
    if isinstance(instance, Image):
        instance.thumbnail_set.all().delete()


@receiver(models.signals.post_delete)
def remove_image(sender, instance, **kwargs):
    if isinstance(instance, Image):
        if not TINT_KEEP_IMAGES:
            instance.image.delete(save=False)

    if isinstance(instance, Thumbnail):
        if not TINT_KEEP_THUMBNAILS:
            instance.image.delete(save=False)



#! /usr/bin/env python
# -*- coding: utf-8 -*-

from setuptools import setup, find_packages
import os

os.environ['DJANGO_SETTINGS_MODULE'] = 'tests.settings'

CLASSIFIERS = [
    'Development Status :: 5 - Production/Stable',
    'Environment :: Web Environment',
    'Framework :: Django',
    'Intended Audience :: Developers',
    'License :: OSI Approved :: BSD License',
    'Operating System :: OS Independent',
    'Programming Language :: Python',
    'Programming Language :: Python :: 2.5',
    'Programming Language :: Python :: 2.6',
    'Programming Language :: Python :: 2.7',
    'Topic :: Internet :: WWW/HTTP',
    'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
    'Topic :: Software Development :: Libraries :: Application Frameworks',
    'Topic :: Software Development :: Libraries :: Python Modules',
]

REQUIREMENTS = [
    'Django >= 1.3',
    'pillow >= 1.7.8',
]

setup(name='django-tint',
      author='Jesús Espino',
      author_email='jespinog@gmail.com',
      description='Transparent image transformation system',
      license='BSD',
      version=':versiontools:tint:',
      setup_requires = [
          'versiontools >= 1.8',
      ],
      test_suite = 'nose.collector',
      tests_require = ['nose >= 1.2.1', 'django >= 1.3.0'],
      packages=find_packages(),
      include_package_data=True,
      classifiers=CLASSIFIERS,
      install_requires=REQUIREMENTS,
      platforms=['any'])



from PIL import Image, ImageOps, ImageEnhance, ImageFile
from tint.imageprocs.default import DefaultImageProc

import unittest
import os

class DefaultImageProcTestCase(unittest.TestCase):
    def assertImageEdges(self, image, edge_width, colors_top, colors_middle, colors_bottom):
        self.assertTrue(self.verify_edges(image, edge_width, colors_top, colors_middle, colors_bottom))

    def assertNotImageEdges(self, image, edge_width, colors_top, colors_middle, colors_bottom):
        self.assertFalse(self.verify_edges(image, edge_width, colors_top, colors_middle, colors_bottom))

    def verify_edges(self, image, edge_width, colors_top=None, colors_middle=None, colors_bottom=None):
        result = True
        width, height = image.size

        for x in range(edge_width):
            if colors_top:
                # TOP check
                result = result and image.getpixel((x, x)) == colors_top[0]
                result = result and image.getpixel((int(width/2), x)) == colors_top[1]
                result = result and image.getpixel((width-edge_width+x, edge_width-x)) == colors_top[2]

            if colors_middle:
                # MIDDLE check
                result = result and image.getpixel((x, int(height/2))) == colors_middle[0]
                result = result and image.getpixel(((width/2), int(height/2))) == colors_middle[1]
                result = result and image.getpixel((width-edge_width+x, int(height/2))) == colors_middle[2]

            if colors_bottom:
                # BOTTOM check
                result = result and image.getpixel((edge_width-x, height-edge_width+x)) == colors_bottom[0]
                result = result and image.getpixel((int(width/2), height-edge_width+x)) == colors_bottom[1]
                result = result and image.getpixel((width-edge_width+x, height-edge_width+x)) == colors_bottom[2]
        return result

    @classmethod
    def setUpClass(cls):
        cls.proc = DefaultImageProc()
        cls.black = (0, 0, 0, 255)
        cls.white = (255, 255, 255, 255)
        cls.red = (255, 0, 0, 255)
        cls.green = (0, 255, 0, 255)
        cls.blue = (0, 0, 255, 255)
        cls.yellow = (255, 255, 0, 255)
        cls.image = Image.open(os.path.join(os.path.dirname(__file__), 'test-image.png'))

    def test_crop(self):
        self.assertRaises(
                Exception,
                lambda: self.proc.crop(self.image.copy(), { "width": 320, "height": 240, "align": 'invalid', "valign": 'top' })
        )

        self.assertRaises(
            Exception,
            lambda: self.proc.crop(self.image.copy(), { "width": 320, "height": 240, "align": 'left', "valign": 'invalid' })
        )

        image_result = self.proc.crop(self.image.copy(), { "width": 1280, "height": 960 })
        self.assertEqual((640, 480), image_result.size)

        image_result = self.proc.crop(self.image.copy(), { "width": 320, "height": 240, "align": 'left', "valign": 'top' })
        self.assertImageEdges(
                image_result,
                10,
                [self.red, self.red, self.white],
                [self.red, self.white, self.white],
                [self.white, self.white, self.white]
        )

        image_result = self.proc.crop(self.image.copy(), { "width": 480, "height": 320, "align": 'center', "valign": 'top' })
        self.assertImageEdges(
                image_result,
                10,
                [self.red, self.white, self.green],
                [self.white, self.white, self.white],
                [self.white, self.white, self.white]
        )

        image_result = self.proc.crop(self.image.copy(), { "width": 320, "height": 240, "align": 'right', "valign": 'top' })
        self.assertImageEdges(
                image_result,
                10,
                [self.white, self.green, self.green],
                [self.white, self.white, self.green],
                [self.white, self.white, self.white]
        )
        image_result = self.proc.crop(self.image.copy(), { "width": 320, "height": 360, "align": 'left', "valign": 'middle' })
        self.assertImageEdges(
                image_result,
                10,
                [self.red, self.white, self.white],
                [self.white, self.white, self.white],
                [self.blue, self.white, self.white]
        )

        image_result = self.proc.crop(self.image.copy(), { "width": 320, "height": 240, "align": 'center', "valign": 'middle' })
        self.assertImageEdges(
                image_result,
                10,
                [self.white, self.white, self.white],
                [self.white, self.white, self.white],
                [self.white, self.white, self.white]
        )

        image_result = self.proc.crop(self.image.copy(), { "width": 320, "height": 360, "align": 'right', "valign": 'middle' })
        self.assertImageEdges(
                image_result,
                10,
                [self.white, self.white, self.green],
                [self.white, self.white, self.white],
                [self.white, self.white, self.yellow]
        )
        image_result = self.proc.crop(self.image.copy(), { "width": 320, "height": 240, "align": 'left', "valign": 'bottom' })
        self.assertImageEdges(
                image_result,
                10,
                [self.white, self.white, self.white],
                [self.blue, self.white, self.white],
                [self.blue, self.blue, self.white]
        )

        image_result = self.proc.crop(self.image.copy(), { "width": 480, "height": 320, "align": 'center', "valign": 'bottom' })
        self.assertImageEdges(
                image_result,
                10,
                [self.white, self.white, self.white],
                [self.white, self.white, self.white],
                [self.blue, self.white, self.yellow]
        )

        image_result = self.proc.crop(self.image.copy(), { "width": 320, "height": 240, "align": 'right', "valign": 'bottom' })
        self.assertImageEdges(
                image_result,
                10,
                [self.white, self.white, self.white],
                [self.white, self.white, self.yellow],
                [self.white, self.yellow, self.yellow]
        )

    def test_mirror(self):
        image_result = self.proc.mirror(self.image.copy(), {})
        self.assertImageEdges(
                image_result,
                10,
                [self.green, self.white, self.red],
                [self.white, self.white, self.white],
                [self.yellow, self.white, self.blue]
        )

    def test_flip(self):
        image_result = self.proc.flip(self.image.copy(), {})
        self.assertImageEdges(
                image_result,
                10,
                [self.blue, self.white, self.yellow],
                [self.white, self.white, self.white],
                [self.red, self.white, self.green]
        )

    def test_grayscale(self):
        image_result = self.proc.grayscale(self.image.copy(), {})
        # Asserting primary colors converted to grayscale
        self.assertImageEdges(
                image_result,
                10,
                [76, 255, 149],
                [255, 255, 255],
                [29, 255, 225]
        )

    def test_invert(self):
        image_result = self.proc.invert(self.image.copy(), {})
        self.assertImageEdges(
                image_result,
                10,
                [(0, 255, 255), self.black[0:-1], (255, 0, 255)],
                [self.black[0:-1], self.black[0:-1], self.black[0:-1]],
                [(255, 255, 0), self.black[0:-1], (0, 0, 255)]
        )

    def test_scale(self):
        image_result = self.proc.scale(self.image.copy(), { "height": 80, "width": 60})
        self.assertImageEdges(
                image_result,
                3,
                [self.red, self.white, self.green],
                [self.white, self.white, self.white],
                [self.blue, self.white, self.yellow]
        )
        self.assertNotImageEdges(
                image_result,
                10,
                [self.red, self.white, self.green],
                [self.white, self.white, self.white],
                [self.blue, self.white, self.yellow]
        )

        image_result = self.proc.scale(self.image.copy(), { "height": 640, "width": 120})
        self.assertEqual(image_result.getpixel((10, 20)), (255, 0, 0, 255))
        self.assertEqual(image_result.getpixel((10, 120)), (255, 255, 255, 255))
        self.assertEqual(image_result.getpixel((3, 120)), (255, 0, 0, 255))

    def test_watermark(self):
        watermark_path = os.path.join(os.path.dirname(__file__), 'watermark.png')
        image_result = self.proc.watermark(self.image.copy(), {"watermark_image": watermark_path, "opacity": 0.5})
        self.assertImageEdges(
                image_result,
                1,
                None,
                [self.white, (128, 128, 128, 255), self.white],
                None,
        )
        image_result = self.proc.watermark(self.image.copy(), {"watermark_image": watermark_path, "opacity": 0})
        self.assertImageEdges(
                image_result,
                1,
                None,
                [self.white, self.white, self.white],
                None,
        )
        image_result = self.proc.watermark(self.image.copy(), {"watermark_image": watermark_path, "opacity": 1})
        self.assertImageEdges(
                image_result,
                1,
                None,
                [self.white, self.black, self.white],
                None,
        )



from tint import settings
from django.utils.importlib import import_module

__all__ = ('image_proc',)

module_name = ".".join(settings.TINT_IMAGE_PROCESSOR.split(".")[0:-1])
class_name = settings.TINT_IMAGE_PROCESSOR.split(".")[-1]
image_proc = getattr(import_module(module_name), class_name)()



from io import BytesIO

from tint.settings import TINT_IMAGE_OUTPUT_FORMAT

from PIL import Image, ImageOps, ImageEnhance, ImageFile


class DefaultImageProc(object):
    def crop(self, image, params):
        (width, height) = image.size

        align = params.get('align', 'center')
        valign = params.get('valign', 'middle')

        if width >= params['width']:
            if align == 'center':
                left = (width/2)-(params['width']/2)
                right = (width/2)+(params['width']/2)
            elif align == 'left':
                left = 0
                right = params['width']
            elif align == 'right':
                left = width-params['width']
                right = width
            else:
                raise Exception('Invalid align')
        else:
            left = 0
            right = width

        if height >= params['height']:
            if valign == 'middle':
                top = (height/2)-(params['height']/2)
                bottom = (height/2)+(params['height']/2)
            elif valign == 'top':
                top = 0
                bottom = params['height']
            elif valign == 'bottom':
                top = height-params['height']
                bottom = height
            else:
                raise Exception('Invalid valign')
        else:
            top = 0
            bottom = height
        return image.crop((left, top, right, bottom))

    def fit(self, image, params):
        num_align = [0.0, 0.0]
        align = params.get('align', 'center')
        valign = params.get('valign', 'middle')

        if align == 'center':
            num_align[1] = 0.5
        elif align == 'left':
            num_align[1] = 0.0
        elif align == 'right':
            num_align[1] = 1.0
        else:
            raise Exception('Invalid align')

        if valign == 'middle':
            num_align[0] = 0.5
        elif valign == 'top':
            num_align[0] = 0.0
        elif valign == 'bottom':
            num_align[0] = 1.0
        else:
            raise Exception('Invalid valign')
        return ImageOps.fit(image, (params['width'], params['height']), Image.ANTIALIAS, 0, num_align)

    def grayscale(self, image, params):
        return ImageOps.grayscale(image)

    def invert(self, image, params):
        image = image.convert('RGB')
        return ImageOps.invert(image)

    def flip(self, image, params):
        return ImageOps.flip(image)

    def mirror(self, image, params):
        return ImageOps.mirror(image)

    def autocontrast(self, image, params):
        return ImageOps.autocontrast(image, params.get('cutoff', 0))

    def equalize(self, image, params):
        return ImageOps.equalize(image)

    def scale(self, image, params):
        return image.resize((params['width'], params['height']), Image.ANTIALIAS)

    def convert(self, image, params):
        return image.convert(mode=params.get('mode', None))

    def watermark(self, image, params):
        orig_watermark = Image.open(params['watermark_image']).convert('RGBA')
        orig_watermark = orig_watermark.resize(image.size, Image.ANTIALIAS)

        watermark = orig_watermark.copy()

        opacity = params.get('opacity', None)
        if opacity is not None:
            alpha = watermark.split()[3]
            alpha = ImageEnhance.Brightness(alpha).enhance(float(opacity))
            watermark.putalpha(alpha)
        return Image.composite(orig_watermark, image, watermark)

    def transform(self, image, *actions):
        image.open()
        im = Image.open(image)
        im_format, im_info = im.format, im.info
        for action in actions:
            im = getattr(self, action['action'])(im, action)

        output_image_format = im_format
        if TINT_IMAGE_OUTPUT_FORMAT:
            output_image_format = TINT_IMAGE_OUTPUT_FORMAT

        # save to memory
        buf = BytesIO()
        try:
            im.save(buf, output_image_format, **im_info)
        except IOError:
            if im.info.get('progression'):
                orig_MAXBLOCK = ImageFile.MAXBLOCK
                temp_MAXBLOCK = 1048576
                if orig_MAXBLOCK >= temp_MAXBLOCK:
                    raise
                ImageFile.MAXBLOCK = temp_MAXBLOCK
                try:
                    im.save(buf, output_image_format, **im_info)
                finally:
                    ImageFile.MAXBLOCK = orig_MAXBLOCK
            else:
                raise
        image.close()

        return buf
